from textwrap import dedent
from unittest import TestCase

from regret import EmittedDeprecation, Deprecator
from regret.testing import Recorder


def calculate():
    """
    Perform a super important calculation.
    """
    return 12


def add(x, y):
    return 12


class TestRegret(TestCase):
    def setUp(self):
        self.recorder = Recorder()
        self.regret = Deprecator(emit=self.recorder.emit)

    def test_function(self):
        self.assertEqual(
            (
                self.regret.callable(version="1.2.3")(calculate)(),
                self.recorder,
            ), (
                12,
                Recorder(saw=[EmittedDeprecation(object=calculate)]),
            ),
        )

    def test_function_with_args(self):
        self.assertEqual(
            (
                self.regret.callable(version="1.2.3")(add)(9, y=3),
                self.recorder,
            ), (
                12,
                Recorder(saw=[EmittedDeprecation(object=add)]),
            ),
        )

    def test_function_gets_deprecation_notice_in_docstring(self):
        deprecated = self.regret.callable(version="v2.3.4")(calculate)
        self.assertEqual(
            deprecated.__doc__, dedent(
                """
                Perform a super important calculation.

                .. deprecated:: v2.3.4
                """,
            )
        )

    def test_function_with_no_docstring_does_not_get_deprecation_notice(self):
        """
        If you're too lazy to add docstrings I ain't helping you.
        """
        def calculate():
            return 12
        deprecated = self.regret.callable(version="v2.3.4")(calculate)
        self.assertIsNone(deprecated.__doc__)

    def test_function_with_replacement(self):
        self.assertEqual(
            (
                self.regret.callable(
                    version="1.2.3",
                    replacement=add,
                )(calculate)(),
                self.recorder,
            ), (
                12,
                Recorder(
                    saw=[
                        EmittedDeprecation(object=calculate, replacement=add),
                    ],
                ),
            ),
        )

    def test_function_is_wrapped(self):
        deprecated = self.regret.callable(version="1.2.3")(calculate)
        self.assertEqual(calculate.__name__, deprecated.__name__)

    def test_original_functions_are_not_mutated(self):
        """
        Deprecating a function in one spot does not mutate the original.

        Any existing references are unchanged.
        """

        def original():
            "Original function docstring."

        original.something = 12
        self.assertEqual(
            (
                original.__name__,
                original.__doc__,
                getattr(original, "__dict__", {}),
            ), (
                "original",
                "Original function docstring.",
                {"something": 12},
            ),
        )

        self.regret.callable(version="1.2.3")(original)

        self.assertEqual(
            (
                original.__name__,
                original.__doc__,
                getattr(original, "__dict__", {}),
            ), (
                "original",
                "Original function docstring.",
                {"something": 12},
            ),
        )

    def test_method(self):
        class Calculator(object):
            def _calculate(self):
                return 12

            calculate = self.regret.callable(version="1.2.3")(_calculate)

        unbound = getattr(
            Calculator._calculate, "im_func", Calculator._calculate,
        )

        self.assertEqual(
            (Calculator().calculate(), self.recorder), (
                12,
                Recorder(saw=[EmittedDeprecation(object=unbound)]),
            ),
        )

    def test_method_is_wrapped(self):
        class Calculator(object):
            def _calculate(self):
                """
                Perform a super important calculation.
                """
                return 12

            calculate = self.regret.callable(version="1.2.3")(_calculate)

        self.assertEqual(
            (
                Calculator.calculate.__name__,
                Calculator().calculate.__name__,
            ),
            (
                Calculator._calculate.__name__,
                Calculator._calculate.__name__,
            ),
        )

    def test_method_gets_deprecation_notice_in_docstring(self):
        class Calculator(object):
            def _calculate(self):
                """
                Perform a super important calculation.
                """
                return 12

            calculate = self.regret.callable(version="4.5.6")(_calculate)

        expected = """
        Perform a super important calculation.

        .. deprecated:: 4.5.6
        """

        self.assertEqual(
            (
                Calculator.calculate.__doc__,
                Calculator().calculate.__doc__,
            ),
            (
                dedent(expected),
                dedent(expected),
            ),
        )

    def test_original_methods_are_not_mutated(self):
        """
        Deprecating a method in one spot does not mutate the original.

        Any existing references are unchanged.
        """

        class Class(object):
            def method(self):
                "Original method docstring."

            method.something = 12

            deprecated = self.regret.callable(version="4.5.6")(method)

        self.assertEqual(
            (
                Class.method.__name__,
                Class.method.__doc__,
                getattr(Class.method, "__dict__", {}),
            ), (
                "method",
                "Original method docstring.",
                {"something": 12},
            ),
        )

    def test_dunder_call(self):
        class Calculator(object):
            def _calculate(self):
                return 12

            __call__ = self.regret.callable(version="1.2.3")(_calculate)

        unbound = getattr(
            Calculator._calculate, "im_func", Calculator._calculate,
        )

        self.assertEqual(
            (Calculator()(), self.recorder), (
                12,
                Recorder(saw=[EmittedDeprecation(object=unbound)]),
            ),
        )



from textwrap import dedent


def doc_with_deprecated_directive(object, name_of, version):
    """
    Add a `deprecated` directive to the provided object's docstring.

    Suitable for use with `regret.Deprecator`.
    """

    return dedent(object.__doc__) + "\n.. deprecated:: {}\n".format(version)



import warnings

from qualname import qualname
from twisted.trial.unittest import SynchronousTestCase

import regret


class Calculator(object):
    @regret.callable(version="1.2.3")
    def __call__(self):
        return 12

    @regret.callable(version="1.2.3")
    def calculate(self):
        return 12


@regret.callable(version="1.2.3")
def calculate():
    return 12


@regret.callable(version="1.2.3")
def add(x, y):
    return x + y


@regret.callable(version="1.2.3", replacement=Calculator)
def calculator_fn():
    return 9


class TestRegret(SynchronousTestCase):
    def assertDeprecated(self, message, filename, fn, args=(), kwargs={}):
        # Sigh... assertWarns takes positional args positionally, instead of as
        # a sequence.
        return self.assertWarns(
            DeprecationWarning,
            message,
            filename,
            fn,
            *args,
            **kwargs
        )

    def test_function(self):
        result = self.assertDeprecated(
            message="'calculate' is deprecated.",
            filename=__file__,
            fn=calculate,
        )
        self.assertEqual(result, 12)

    def test_function_with_args(self):
        result = self.assertDeprecated(
            message="'add' is deprecated.",
            filename=__file__,
            fn=add,
            args=(9,),
            kwargs=dict(y=3),
        )
        self.assertEqual(result, 12)

    def test_function_with_replacement(self):
        result = self.assertDeprecated(
            message=(
                "'calculator_fn' is deprecated. "
                "Please use 'Calculator' instead."
            ),
            filename=__file__,
            fn=calculator_fn,
        )
        self.assertEqual(result, 9)

    def test_method(self):
        calculator = Calculator()
        result = self.assertDeprecated(
            message="'Calculator.calculate' is deprecated.",
            filename=__file__,
            fn=calculator.calculate,
        )
        self.assertEqual(result, 12)

    def test_dunder_call(self):
        calculator = Calculator()
        result = self.assertDeprecated(
            message="'Calculator.__call__' is deprecated.",
            filename=__file__,
            fn=calculator,
        )
        self.assertEqual(result, 12)

    def test_custom_qualname(self):
        etacerped = regret.Deprecator(
            name_of=lambda object: object.__name__[::-1],
        )
        @etacerped.callable(version="v1.2.3")
        def calculate():
            return 12

        result = self.assertDeprecated(
            message="'etaluclac' is deprecated.",
            filename=__file__,
            fn=calculate,
        )

    def test_custom_docstring_modifier(self):
        deprecate = regret.Deprecator(
            name_of=lambda object: "OBJECTNAME",
            new_docstring=lambda object, name_of, version: (
                object.__doc__ + name_of(object) + " deprecated in " + version
            )
        )
        @deprecate.callable(version="v1.2.3")
        def calculate():
            "Very important docstring for "

        self.assertEqual(
            calculate.__doc__,
            "Very important docstring for OBJECTNAME deprecated in v1.2.3",
        )



from functools import wraps

from qualname import qualname
import attr

from regret import _sphinx, _warnings


@attr.s(eq=True, frozen=True)
class Deprecator(object):
    """
    Deprecators help manifest regret.

    Arguments:

        emit:

            a callable which will be called with one argument, an
            `EmittedDeprecation` instance, whenever a deprecated object
            has been used. If unprovided, by default, a warning will be
            shown using the standard library `warnings` module.

        name_of:

            a callable which given any Python object should
            return a suitable name for the object. If unprovided,
            `qualname.qualname` will be used, and therefore an object's
            (non-fully-)qualified name will appear in messages.

        new_docstring:

            a callable which should produce a docstring for newly
            deprecated objects. It will be called with three *keyword*
            arguments:

                * ``object``, the object that is being depreated

                * ``name_of``, the callable described above for use in
                  calculating object names

                * ``version``, the version that deprecates the provided object

            and it should return a single string which will become the new
            docstring for a deprecated object. If unprovided, deprecation
            docstrings will be constructed using syntax suitable for `Sphinx`,
            via the `deprecated` directive.
    """

    _emit = attr.ib(default=_warnings.emit)
    _name_of = attr.ib(default=qualname)
    _new_docstring = attr.ib(default=_sphinx.doc_with_deprecated_directive)

    def emit_deprecation(self, **kwargs):
        self._emit(EmittedDeprecation(name_of=self._name_of, **kwargs))

    # -- Deprecatable objects --

    def callable(self, version, replacement=None):
        """
        Deprecate a callable as of the given version.

        Arguments:

            version:

                the first version in which the deprecated object was considered
                deprecated

            replacement:

                optionally, an object that is the (direct or indirect)
                replacement for the functionality previously performed
                by the deprecated callable
        """

        def deprecate(thing):
            @wraps(thing)
            def call_deprecated(*args, **kwargs):
                self.emit_deprecation(object=thing, replacement=replacement)
                return thing(*args, **kwargs)

            __doc__ = thing.__doc__
            if __doc__ is not None:
                call_deprecated.__doc__ = self._new_docstring(
                    object=thing,
                    name_of=self._name_of,
                    version=version,
                )

            return call_deprecated
        return deprecate


@attr.s(eq=True, frozen=True, hash=True)
class EmittedDeprecation(object):

    _object = attr.ib()
    _name_of = attr.ib(default=qualname)
    _replacement = attr.ib(default=None)

    def message(self):
        if self._replacement is None:
            replacement_info = ""
        else:
            replacement_info = " Please use {!r} instead.".format(
                self._name_of(self._replacement),
            )
        return "{name!r} is deprecated.{replacement_info}".format(
            name=self._name_of(self._object),
            replacement_info=replacement_info,
        )


_DEPRECATOR = Deprecator()
